# MIT License
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
"""
Implementation of CRC-32C taken from translating the rust code we are using
to python.  As rust crc code is MIT licensed, so shall this be too.
"""

_CASTAGNOLI_TABLE = (
    0x00000000,
    0xF26B8303,
    0xE13B70F7,
    0x1350F3F4,
    0xC79A971F,
    0x35F1141C,
    0x26A1E7E8,
    0xD4CA64EB,
    0x8AD958CF,
    0x78B2DBCC,
    0x6BE22838,
    0x9989AB3B,
    0x4D43CFD0,
    0xBF284CD3,
    0xAC78BF27,
    0x5E133C24,
    0x105EC76F,
    0xE235446C,
    0xF165B798,
    0x030E349B,
    0xD7C45070,
    0x25AFD373,
    0x36FF2087,
    0xC494A384,
    0x9A879FA0,
    0x68EC1CA3,
    0x7BBCEF57,
    0x89D76C54,
    0x5D1D08BF,
    0xAF768BBC,
    0xBC267848,
    0x4E4DFB4B,
    0x20BD8EDE,
    0xD2D60DDD,
    0xC186FE29,
    0x33ED7D2A,
    0xE72719C1,
    0x154C9AC2,
    0x061C6936,
    0xF477EA35,
    0xAA64D611,
    0x580F5512,
    0x4B5FA6E6,
    0xB93425E5,
    0x6DFE410E,
    0x9F95C20D,
    0x8CC531F9,
    0x7EAEB2FA,
    0x30E349B1,
    0xC288CAB2,
    0xD1D83946,
    0x23B3BA45,
    0xF779DEAE,
    0x05125DAD,
    0x1642AE59,
    0xE4292D5A,
    0xBA3A117E,
    0x4851927D,
    0x5B016189,
    0xA96AE28A,
    0x7DA08661,
    0x8FCB0562,
    0x9C9BF696,
    0x6EF07595,
    0x417B1DBC,
    0xB3109EBF,
    0xA0406D4B,
    0x522BEE48,
    0x86E18AA3,
    0x748A09A0,
    0x67DAFA54,
    0x95B17957,
    0xCBA24573,
    0x39C9C670,
    0x2A993584,
    0xD8F2B687,
    0x0C38D26C,
    0xFE53516F,
    0xED03A29B,
    0x1F682198,
    0x5125DAD3,
    0xA34E59D0,
    0xB01EAA24,
    0x42752927,
    0x96BF4DCC,
    0x64D4CECF,
    0x77843D3B,
    0x85EFBE38,
    0xDBFC821C,
    0x2997011F,
    0x3AC7F2EB,
    0xC8AC71E8,
    0x1C661503,
    0xEE0D9600,
    0xFD5D65F4,
    0x0F36E6F7,
    0x61C69362,
    0x93AD1061,
    0x80FDE395,
    0x72966096,
    0xA65C047D,
    0x5437877E,
    0x4767748A,
    0xB50CF789,
    0xEB1FCBAD,
    0x197448AE,
    0x0A24BB5A,
    0xF84F3859,
    0x2C855CB2,
    0xDEEEDFB1,
    0xCDBE2C45,
    0x3FD5AF46,
    0x7198540D,
    0x83F3D70E,
    0x90A324FA,
    0x62C8A7F9,
    0xB602C312,
    0x44694011,
    0x5739B3E5,
    0xA55230E6,
    0xFB410CC2,
    0x092A8FC1,
    0x1A7A7C35,
    0xE811FF36,
    0x3CDB9BDD,
    0xCEB018DE,
    0xDDE0EB2A,
    0x2F8B6829,
    0x82F63B78,
    0x709DB87B,
    0x63CD4B8F,
    0x91A6C88C,
    0x456CAC67,
    0xB7072F64,
    0xA457DC90,
    0x563C5F93,
    0x082F63B7,
    0xFA44E0B4,
    0xE9141340,
    0x1B7F9043,
    0xCFB5F4A8,
    0x3DDE77AB,
    0x2E8E845F,
    0xDCE5075C,
    0x92A8FC17,
    0x60C37F14,
    0x73938CE0,
    0x81F80FE3,
    0x55326B08,
    0xA759E80B,
    0xB4091BFF,
    0x466298FC,
    0x1871A4D8,
    0xEA1A27DB,
    0xF94AD42F,
    0x0B21572C,
    0xDFEB33C7,
    0x2D80B0C4,
    0x3ED04330,
    0xCCBBC033,
    0xA24BB5A6,
    0x502036A5,
    0x4370C551,
    0xB11B4652,
    0x65D122B9,
    0x97BAA1BA,
    0x84EA524E,
    0x7681D14D,
    0x2892ED69,
    0xDAF96E6A,
    0xC9A99D9E,
    0x3BC21E9D,
    0xEF087A76,
    0x1D63F975,
    0x0E330A81,
    0xFC588982,
    0xB21572C9,
    0x407EF1CA,
    0x532E023E,
    0xA145813D,
    0x758FE5D6,
    0x87E466D5,
    0x94B49521,
    0x66DF1622,
    0x38CC2A06,
    0xCAA7A905,
    0xD9F75AF1,
    0x2B9CD9F2,
    0xFF56BD19,
    0x0D3D3E1A,
    0x1E6DCDEE,
    0xEC064EED,
    0xC38D26C4,
    0x31E6A5C7,
    0x22B65633,
    0xD0DDD530,
    0x0417B1DB,
    0xF67C32D8,
    0xE52CC12C,
    0x1747422F,
    0x49547E0B,
    0xBB3FFD08,
    0xA86F0EFC,
    0x5A048DFF,
    0x8ECEE914,
    0x7CA56A17,
    0x6FF599E3,
    0x9D9E1AE0,
    0xD3D3E1AB,
    0x21B862A8,
    0x32E8915C,
    0xC083125F,
    0x144976B4,
    0xE622F5B7,
    0xF5720643,
    0x07198540,
    0x590AB964,
    0xAB613A67,
    0xB831C993,
    0x4A5A4A90,
    0x9E902E7B,
    0x6CFBAD78,
    0x7FAB5E8C,
    0x8DC0DD8F,
    0xE330A81A,
    0x115B2B19,
    0x020BD8ED,
    0xF0605BEE,
    0x24AA3F05,
    0xD6C1BC06,
    0xC5914FF2,
    0x37FACCF1,
    0x69E9F0D5,
    0x9B8273D6,
    0x88D28022,
    0x7AB90321,
    0xAE7367CA,
    0x5C18E4C9,
    0x4F48173D,
    0xBD23943E,
    0xF36E6F75,
    0x0105EC76,
    0x12551F82,
    0xE03E9C81,
    0x34F4F86A,
    0xC69F7B69,
    0xD5CF889D,
    0x27A40B9E,
    0x79B737BA,
    0x8BDCB4B9,
    0x988C474D,
    0x6AE7C44E,
    0xBE2DA0A5,
    0x4C4623A6,
    0x5F16D052,
    0xAD7D5351,
)


def crc(buf):
    """
    Calculate CRC32C aka. Castagnoli
    :param buf:  Byte array with data to crc
    :return: crc32c
    """
    crc_val = 0xFFFFFFFF

    for byte in buf:
        tbl_idx = (crc_val ^ byte) & 0xFF
        crc_val = (_CASTAGNOLI_TABLE[tbl_idx] ^ (crc_val >> 8)) & 0xFFFFFFFF

    return ~crc_val & 0xFFFFFFFF



#!/usr/bin/python3
#
# Copyright 2020 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Uploads the vendored dependency tarball as a release asset.
"""

# isort: STDLIB
import os
import re
import sys
from getpass import getpass

# isort: THIRDPARTY
from github import Github


def main():
    """
    Main function
    """

    if len(sys.argv) < 2:
        print("USAGE: %s <RELEASE_VERSION>" % sys.argv[0])
        print("\tRELEASE_VERSION: MAJOR.MINOR.PATCH")
        raise RuntimeError("One positional argument is required")

    release_version = sys.argv[1]
    if re.match(r"^[0-9]+\.[0-9]+\.[0-9]$", release_version) is None:
        raise RuntimeError("Invalid release version %s provided" % release_version)

    api_key = os.environ.get("GITHUB_API_KEY")
    if api_key is None:
        api_key = getpass("API key: ")

    git = Github(api_key)

    repo = git.get_repo("stratis-storage/stratisd")

    release = repo.create_git_release(
        "v%s" % release_version,
        "Version %s" % release_version,
        "See changelog here: https://github.com/stratis-storage/stratisd/blob/master/CHANGES.txt",
        draft=True,
    )

    label = "stratisd-%s-vendor.tar.gz" % release_version
    release.upload_asset(label, label)


if __name__ == "__main__":
    try:
        main()
    except Exception as err:  # pylint: disable=broad-except
        print(err)
        sys.exit(1)



#!/usr/bin/python3
#
# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Identify a device as a Stratis block device.
"""
# isort: STDLIB
import struct
import sys
from collections import namedtuple

# isort: THIRDPARTY
from _crc32 import crc

BS = 512
FIRST_COPY_OFFSET = BS
SECOND_COPY_OFFSET = BS * 9
SB_AREA_SIZE = 16 * BS
STRATIS_MAGIC = b"!Stra0tis\x86\xff\x02^Arh"

MAGIC_OFFSET = 4
MAGIC_LEN = len(STRATIS_MAGIC)


def _valid_stratis_sb(buf):
    """
    Check to see if the buffer is a valid Stratis super block
    :param buf: Byte buffer starting at Stratis block offset
    :return: None or named tuple
    """
    if buf[MAGIC_OFFSET : MAGIC_OFFSET + MAGIC_LEN] == STRATIS_MAGIC:
        # Verify CRC
        if crc(buf[MAGIC_OFFSET:BS]) == struct.unpack_from("<L", buf, 0)[0]:
            super_block = namedtuple(
                "StratisSuperblock",
                "CRC32C STRATIS_MAGIC_NUMBER BLKDEV_SIZE_SECTORS "
                "SIGBLOCK_VERSION UNUSED POOL_UUID DEV_UUID MDA_SIZE_SECTORS "
                "RESERVED_SIZE_SECTORS FLAGS INITIALIZATION_TIME_SECONDS",
            )

            return super_block._make(struct.unpack_from("<L16sQc3s32s32sQQQQ", buf))
    return None


def stratis_signature(block_device):
    """
    Checks a device to see if it has a valid Stratis signature on it.
    :param block_device:
    :return: None if not Stratis, else named tuple
    """
    try:
        with open(block_device, "r+b") as header:
            buf = header.read(SB_AREA_SIZE)
    # pylint: disable=bare-except
    except:
        return None

    return _valid_stratis_sb(buf[FIRST_COPY_OFFSET:]) or _valid_stratis_sb(
        buf[SECOND_COPY_OFFSET:]
    )


def main():
    """
    The main method
    """
    if len(sys.argv) != 2:
        print("syntax: stratis_signature.py <block device>")
        sys.exit(2)

    sig = stratis_signature(sys.argv[1])
    if not sig:
        sys.exit(1)

    print(sig)
    sys.exit(0)


if __name__ == "__main__":
    main()



# isort: STDLIB
import os
import sys

# isort: THIRDPARTY
import setuptools

if sys.version_info[0] < 3:
    from codecs import open


def local_file(name):
    return os.path.relpath(os.path.join(os.path.dirname(__file__), name))


README = local_file("README.rst")

setuptools.setup(
    name="stratisd-client-dbus",
    author="Anne Mulhern",
    author_email="amulhern@redhat.com",
    description="testing library for stratisd",
    long_description=open(README, encoding="utf-8").read(),
    platforms=["Linux"],
    install_requires=["dbus-client-gen>=0.3", "dbus-python-client-gen>=0.6"],
    package_dir={"": "src"},
    packages=setuptools.find_packages("src"),
)



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Functionality pertaining to device mapper
"""

# isort: STDLIB
import os
import subprocess

_DM_BIN = os.getenv("STRATIS_DMSETUP_BIN", "/usr/sbin/dmsetup")


def _get_stratis_devices():
    """
    Return a list of stratis DM devices
    :return: A list of stratis DM devices
    """
    output = subprocess.check_output([_DM_BIN, "ls"])
    decoded = output.decode("utf-8")
    return [l for l in decoded.split() if l.startswith("stratis-")]


def _remove_device(device):
    """
    Attempt to remove a device mapper device.
    :param device: The device to remove
    :return: True if device was removed, else False
    """
    return_code = subprocess.call([_DM_BIN, "remove", device], stderr=subprocess.PIPE)
    return return_code == 0


def remove_stratis_setup():
    """
    Removes all device mapper tables which are related to stratis (best attempt)
    :return: None
    """
    attempts = 0
    devices = _get_stratis_devices()
    max_loops = len(devices) * 5

    # There is some dependency ordering with regards to dm tables, we will
    # simply iterate over the list attempting to remove them until they are
    # all gone or we give up trying.
    while devices != [] and attempts < max_loops:
        attempts += 1

        # Make a copy of the list as we are traversing a list we want to
        # modify.
        for dev in devices.copy():
            if _remove_device(dev):
                devices.remove(dev)


if __name__ == "__main__":
    remove_stratis_setup()



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Used to test behavior of the udev device discovery mechanism.
"""

# isort: STDLIB
import random

# isort: LOCAL
from stratisd_client_dbus import (
    FetchProperties,
    Manager,
    Pool,
    StratisdErrors,
    get_object,
)
from stratisd_client_dbus._constants import TOP_OBJECT
from stratisd_client_dbus._stratisd_constants import EncryptionMethod

from ._loopback import UDEV_ADD_EVENT
from ._utils import (
    CRYPTO_LUKS_FS_TYPE,
    STRATIS_FS_TYPE,
    OptionalKeyServiceContextManager,
    ServiceContextManager,
    UdevTest,
    create_pool,
    get_devnodes,
    random_string,
    remove_stratis_dm_devices,
    settle,
    wait_for_udev,
    wait_for_udev_count,
)

LOCKED_POOL_UUIDS_PROP_NAME = "LockedPoolUuids"


class UdevTest1(UdevTest):
    """
    See description of test in _test_driver method.
    """

    def _test_driver(self, number_of_pools, dev_count_pool):
        """
        Run the following test:

        0. Start stratisd.
        1. Create number_of_pools pools each with dev_count_pool devices.
        2. Stop stratisd and take down all Stratis dm devices.
        3. Verify that the number of devices with Stratis metadata is the
        same as the number of devices used when creating pools.
        4. Start stratisd, verify that it can find the correct number of pools.
        5. Stop stratisd and take down all Stratis dm devices.
        6. Unplug all the loopbacked devices.
        7. Verify that no devices with Stratis metadata can be found.
        8. Start stratisd, verify that no pools are found.
        9. Plug all but the last device for each pool. Verify that stratisd
        reports no pools.
        10. Stop stratisd and restart it. Verify that stratisd reports no pools.
        11. Add the last device for each pool, verify that stratisd detects
        all pools.

        :param int number_of_pools: the number of pools to use in the test
        :param int dev_count_pool: the number of devices per pool
        """

        pool_data = {}
        with ServiceContextManager():
            for _ in range(number_of_pools):
                device_tokens = self._lb_mgr.create_devices(dev_count_pool)

                pool_name = random_string(5)

                create_pool(pool_name, self._lb_mgr.device_files(device_tokens))
                pool_data[pool_name] = device_tokens

        remove_stratis_dm_devices()

        all_tokens = [
            dev for device_tokens in pool_data.values() for dev in device_tokens
        ]
        all_devnodes = self._lb_mgr.device_files(all_tokens)

        wait_for_udev(STRATIS_FS_TYPE, all_devnodes)

        with ServiceContextManager():
            self.wait_for_pools(number_of_pools)

        remove_stratis_dm_devices()

        self._lb_mgr.unplug(all_tokens)

        wait_for_udev(STRATIS_FS_TYPE, [])

        last_index = dev_count_pool - 1
        with ServiceContextManager():
            self.wait_for_pools(0)

            # Add all but the last device for each pool
            tokens_to_add = [
                tok
                for device_tokens in pool_data.values()
                for tok in device_tokens[:last_index]
            ]
            self._lb_mgr.hotplug(tokens_to_add)
            wait_for_udev(STRATIS_FS_TYPE, self._lb_mgr.device_files(tokens_to_add))

            self.wait_for_pools(0)

            # Add the last device that makes each pool complete
            self._lb_mgr.hotplug(
                [device_tokens[last_index] for device_tokens in pool_data.values()]
            )

            wait_for_udev(STRATIS_FS_TYPE, all_devnodes)

            self.wait_for_pools(number_of_pools)

            for name in pool_data:
                self.wait_for_pools(1, name=name)

        remove_stratis_dm_devices()

    def test_generic(self):
        """
        See _test_driver for description.
        """
        self._test_driver(2, 4)


class UdevTest2(UdevTest):
    """
    Exercise a single pool.
    """

    def _single_pool(self, num_devices, *, num_hotplugs=0):
        """
        Creates a single pool with specified number of devices.

        Verifies the following:
        * On service start there are no pools
        * After pool creation there is one pool and all block devices passed
        to the pool creation method have Stratis metadata
        * After the daemon is brought down and restarted it has found a pool
        * After the loop backed devices have been removed no devices with
        Stratis metadata are found and the newly brought up daemon finds 0
        pools.
        * After the devices are re-added, they can all be found with Stratis
        metadata and the daemon has a pool.
        * Causing num_hotplugs synthetic udev events for each device has
        no further effect, i.e., no additional pools suddenly appear.

        :param int num_devices: Number of devices to use for pool
        :param int num_hotplugs: Number of synthetic udev "add" event per device
        :return: None
        """
        device_tokens = self._lb_mgr.create_devices(num_devices)
        devnodes = self._lb_mgr.device_files(device_tokens)

        with ServiceContextManager():
            self.wait_for_pools(0)
            (_, (_, device_object_paths)) = create_pool(random_string(5), devnodes)
            self.wait_for_pools(1)

            self.assertEqual(len(device_object_paths), len(devnodes))
            wait_for_udev(STRATIS_FS_TYPE, get_devnodes(device_object_paths))

        remove_stratis_dm_devices()

        with ServiceContextManager():
            self.wait_for_pools(1)

        remove_stratis_dm_devices()

        self._lb_mgr.unplug(device_tokens)

        wait_for_udev(STRATIS_FS_TYPE, [])

        with ServiceContextManager():
            self.wait_for_pools(0)

            self._lb_mgr.hotplug(device_tokens)

            wait_for_udev(STRATIS_FS_TYPE, devnodes)

            self.wait_for_pools(1)

            for _ in range(num_hotplugs):
                self._lb_mgr.generate_synthetic_udev_events(
                    device_tokens, UDEV_ADD_EVENT
                )

            settle()

            self.wait_for_pools(1)

        remove_stratis_dm_devices()

    def test_simultaneous(self):
        """
        See documentation for _single_pool.
        """
        self._single_pool(8)

    def test_spurious_adds(self):
        """
        See documentation for _single_pool.
        """
        self._single_pool(4, num_hotplugs=4)

    def test_simple_udev_add(self):
        """
        See documentation for _single_pool.
        """
        self._single_pool(1, num_hotplugs=1)


class UdevTest3(UdevTest):
    """
    A very simple test that just creates a pool, and then brings down the
    daemon, brings it up again, and allows it to discover the existing pool.
    """

    def _simple_initial_discovery_test(  # pylint: disable=bad-continuation
        self, *, key_spec=None, take_down_dm=False
    ):  # pylint: disable=too-many-locals
        """
        A simple test of discovery on start up.

        * Create just one pool
        * Stop the daemon
        * Restart the daemon and verify that the pool is found

        :param key_spec: specification for a key to be inserted into the kernel
                         keyring consisting of the key description and key data
        :type key_spec: (str, bytes) or NoneType
        :param bool take_down_dm: if True take down all Stratis devicemapper
        devices once stratisd is shut down
        """
        num_devices = 3
        device_tokens = self._lb_mgr.create_devices(num_devices)
        devnodes = self._lb_mgr.device_files(device_tokens)
        key_spec = None if key_spec is None else [key_spec]

        with OptionalKeyServiceContextManager(key_spec=key_spec) as key_descriptions:
            key_description = None if key_spec is None else key_descriptions[0]

            self.wait_for_pools(0)
            (_, (pool_object_path, device_object_paths)) = create_pool(
                random_string(5), devnodes, key_description=key_description
            )
            pool_uuid = Pool.Properties.Uuid.Get(get_object(pool_object_path))

            self.wait_for_pools(1)

            wait_for_udev(STRATIS_FS_TYPE, get_devnodes(device_object_paths))

        if take_down_dm:
            remove_stratis_dm_devices()

        with OptionalKeyServiceContextManager(key_spec=key_spec):
            ((option, unlock_uuids), exit_code, _) = Manager.Methods.UnlockPool(
                get_object(TOP_OBJECT),
                {
                    "pool_uuid": pool_uuid,
                    "unlock_method": str(EncryptionMethod.KEYRING),
                },
            )
            if key_spec is None:
                self.assertNotEqual(exit_code, StratisdErrors.OK)
                self.assertEqual(option, False)
            else:
                self.assertEqual(exit_code, StratisdErrors.OK)
                self.assertEqual(option, take_down_dm)
                self.assertEqual(len(unlock_uuids), num_devices if take_down_dm else 0)

            wait_for_udev_count(num_devices)

            self.wait_for_pools(1)

        remove_stratis_dm_devices()

    def test_encryption_simple_initial_discovery(self):
        """
        See documentation for _simple_initial_discovery_test.
        """
        self._simple_initial_discovery_test(key_spec=("test_key_desc", "test_key"))

    def test_simple_initial_discovery(self):
        """
        See documentation for _simple_initial_discovery_test.
        """
        self._simple_initial_discovery_test()

    def test_encryption_simple_initial_discovery_with_takedown(self):
        """
        See documentation for _simple_initial_discovery_test.
        """
        self._simple_initial_discovery_test(
            key_spec=("test_key_desc", "test_key"), take_down_dm=True
        )


class UdevTest4(UdevTest):
    """
    A test that verifies successful discovery of devices via udev events.

    A pool is created. Then the daemon is brought down and all Stratis
    devicemapper devices are destroyed and the devices are unplugged.

    Then the daemon is brought back up again. The devices are plugged back
    in, and it is verified that the daemon has recreated the pool.
    """

    def _simple_event_test(self, *, key_spec=None):  # pylint: disable=too-many-locals
        """
        A simple test of event-based discovery.

        * Create just one pool.
        * Stop the daemon.
        * Unplug the devices.
        * Start the daemon.
        * Plug the devices in one by one. The pool should come up when the last
        device is plugged in.

        :param key_spec: specification for a key to be inserted into the kernel
                         keyring consisting of the key description and key data
        :type key_spec: (str, bytes) or NoneType
        """
        num_devices = 3
        udev_wait_type = STRATIS_FS_TYPE if key_spec is None else CRYPTO_LUKS_FS_TYPE
        device_tokens = self._lb_mgr.create_devices(num_devices)
        devnodes = self._lb_mgr.device_files(device_tokens)
        key_spec = None if key_spec is None else [key_spec]

        with OptionalKeyServiceContextManager(key_spec=key_spec) as key_descriptions:
            key_description = None if key_spec is None else key_descriptions[0]

            self.wait_for_pools(0)
            (_, (pool_object_path, _)) = create_pool(
                random_string(5), devnodes, key_description=key_description
            )
            pool_uuid = Pool.Properties.Uuid.Get(get_object(pool_object_path))

            self.wait_for_pools(1)

        remove_stratis_dm_devices()

        self._lb_mgr.unplug(device_tokens)
        wait_for_udev(udev_wait_type, [])

        with OptionalKeyServiceContextManager(key_spec=key_spec):
            self.wait_for_pools(0)

            indices = list(range(num_devices))
            random.shuffle(indices)

            tokens_up = []
            for index in indices[:-1]:
                tokens_up.append(device_tokens[index])
                self._lb_mgr.hotplug([tokens_up[-1]])
                wait_for_udev(udev_wait_type, self._lb_mgr.device_files(tokens_up))
                self.wait_for_pools(0)

            ((option, unlock_uuids), exit_code, _) = Manager.Methods.UnlockPool(
                get_object(TOP_OBJECT),
                {
                    "pool_uuid": pool_uuid,
                    "unlock_method": str(EncryptionMethod.KEYRING),
                },
            )
            if key_spec is None:
                self.assertNotEqual(exit_code, StratisdErrors.OK)
                self.assertEqual(option, False)
            else:
                self.assertEqual(exit_code, StratisdErrors.OK)
                self.assertEqual(option, True)
                self.assertEqual(len(unlock_uuids), num_devices - 1)

            self.wait_for_pools(0)

            tokens_up.append(device_tokens[indices[-1]])
            self._lb_mgr.hotplug([tokens_up[-1]])

            wait_for_udev(udev_wait_type, self._lb_mgr.device_files(tokens_up))

            ((option, unlock_uuids), exit_code, _) = Manager.Methods.UnlockPool(
                get_object(TOP_OBJECT),
                {
                    "pool_uuid": pool_uuid,
                    "unlock_method": str(EncryptionMethod.KEYRING),
                },
            )

            if key_spec is None:
                self.assertNotEqual(exit_code, StratisdErrors.OK)
                self.assertEqual(option, False)
            else:
                self.assertEqual(exit_code, StratisdErrors.OK)
                self.assertEqual(option, True)
                self.assertEqual(len(unlock_uuids), 1)

            wait_for_udev_count(num_devices)

            self.wait_for_pools(1)

        remove_stratis_dm_devices()

    def test_simple_event(self):
        """
        See documentation for _simple_event_test.
        """
        self._simple_event_test()

    def test_encryption_simple_event(self):
        """
        See documentation for _simple_event_test.
        """
        self._simple_event_test(key_spec=("test_key_desc", "test_key"))


class UdevTest5(UdevTest):
    """
    Test correct handling of pools with duplicate pool names.

    This test creates multiple pools with the same name but different UUIDs.
    It is possible to do this by repeatedly bringing up the daemon, creating
    a pool, bringing the daemon down again, and then unplugging the devices
    belonging to that pool. When the daemon comes up again, the previously
    created pool is invisible, and so another of the same name can be created.

    Next, the daemon is brought up and all the previously created devices
    are made visible. Only one pool is set up, the others must all be placed
    in the set of liminal devices, because they represent pools with the same
    name. Then all existing pools are renamed. Then, all devices receive
    synthetic events, which should cause another pool to be discovered, and
    so forth. Eventually, all pools should have been set up.
    """

    def test_duplicate_pool_name(self):  # pylint: disable=too-many-locals
        """
        Create more than one pool with the same name, then dynamically fix it
        :return: None
        """
        pool_name = random_string(12)
        pool_tokens = []
        encrypted_indices = []
        unencrypted_indices = []
        num_pools = 3
        keys = [
            ("key_desc_1", "key_data_1"),
            ("key_desc_2", "key_data_2"),
            ("key_desc_3", "key_data_3"),
        ]

        # Create some pools with duplicate names
        for i in range(num_pools):
            this_pool = self._lb_mgr.create_devices(i + 1)
            devnodes = self._lb_mgr.device_files(this_pool)

            with OptionalKeyServiceContextManager(key_spec=keys) as key_descriptions:
                key_description = (
                    key_descriptions[random.randint(0, len(key_descriptions) - 1)]
                    if random.choice([True, False])
                    else None
                )
                create_pool(pool_name, devnodes, key_description=key_description)
                if key_description is None:
                    unencrypted_indices.append(i)
                else:
                    encrypted_indices.append(i)

            pool_tokens.append(this_pool)

            remove_stratis_dm_devices()

            self._lb_mgr.unplug(this_pool)

            wait_for_udev(STRATIS_FS_TYPE, [])

        all_tokens = [dev for sublist in pool_tokens for dev in sublist]
        random.shuffle(all_tokens)

        with OptionalKeyServiceContextManager(key_spec=keys):
            self._lb_mgr.hotplug(all_tokens)

            (luks_tokens, non_luks_tokens) = (
                [
                    dev
                    for sublist in (pool_tokens[i] for i in encrypted_indices)
                    for dev in sublist
                ],
                [
                    dev
                    for sublist in (pool_tokens[i] for i in unencrypted_indices)
                    for dev in sublist
                ],
            )

            wait_for_udev(CRYPTO_LUKS_FS_TYPE, self._lb_mgr.device_files(luks_tokens))
            wait_for_udev(STRATIS_FS_TYPE, self._lb_mgr.device_files(non_luks_tokens))

            (valid, variant_pool_uuids) = FetchProperties.Methods.GetProperties(
                get_object(TOP_OBJECT), {"properties": [LOCKED_POOL_UUIDS_PROP_NAME]}
            )[LOCKED_POOL_UUIDS_PROP_NAME]

            self.assertTrue(valid)

            for pool_uuid in variant_pool_uuids:
                ((option, _), exit_code, _) = Manager.Methods.UnlockPool(
                    get_object(TOP_OBJECT),
                    {
                        "pool_uuid": pool_uuid,
                        "unlock_method": str(EncryptionMethod.KEYRING),
                    },
                )
                self.assertEqual(exit_code, StratisdErrors.OK)
                self.assertEqual(option, True)

            wait_for_udev_count(len(all_tokens))

            # The number of pools should never exceed one, since all the pools
            # previously formed in the test have the same name.
            self.wait_for_pools(1)

            # Dynamically rename all active pools to a randomly chosen name,
            # then generate synthetic add events for every loopbacked device.
            # After num_pools - 1 iterations, all pools should have been set up.
            for pool_count in range(num_pools - 1):
                current_pools = self.wait_for_pools(pool_count + 1)

                # Rename all active pools to a randomly selected new name
                for object_path, _ in current_pools:
                    Pool.Methods.SetName(
                        get_object(object_path), {"name": random_string(10)}
                    )

                self._lb_mgr.generate_synthetic_udev_events(all_tokens, UDEV_ADD_EVENT)

                settle()

                self.wait_for_pools(len(current_pools) + 1)

            self.wait_for_pools(num_pools)

        remove_stratis_dm_devices()



#!/usr/bin/python

"""
Run lint checks on files in directory.
"""

# isort: STDLIB
import argparse
import subprocess
import sys

ARG_MAP = {
    "check.py": [
        "--reports=no",
        "--disable=I",
        "--msg-template='{path}:{line}: [{msg_id}({symbol}), {obj}] {msg}'",
    ],
    "compare_fedora_versions": [
        "--reports=no",
        "--disable=I",
        "--msg-template='{path}:{line}: [{msg_id}({symbol}), {obj}] {msg}'",
    ],
    "create_release.py": [
        "--reports=no",
        "--disable=I",
        "--msg-template='{path}:{line}: [{msg_id}({symbol}), {obj}] {msg}'",
    ],
    "set_lower_bounds": [
        "--reports=no",
        "--disable=I",
        "--msg-template='{path}:{line}: [{msg_id}({symbol}), {obj}] {msg}'",
    ],
    "_utils.py": [
        "--reports=no",
        "--disable=I",
        "--msg-template='{path}:{line}: [{msg_id}({symbol}), {obj}] {msg}'",
    ],
}


def get_parser():
    """
    Generate an appropriate parser.

    :returns: an argument parser
    :rtype: `ArgumentParser`
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "package", choices=ARG_MAP.keys(), help="designates the package to test"
    )
    return parser


def get_command(namespace):
    """
    Get the pylint command for these arguments.

    :param `Namespace` namespace: the namespace
    """
    return ["pylint", namespace.package] + ARG_MAP[namespace.package]


def main():
    """
    The main entry method
    """
    args = get_parser().parse_args()
    return subprocess.call(get_command(args), stdout=sys.stdout)


if __name__ == "__main__":
    sys.exit(main())
